@@ -1,7 +1,7 @@
 package Mail::Milter::Authentication::Config;
 use strict;
 use warnings;
-use version; our $VERSION = version->declare('v1.0.1');
+use version; our $VERSION = version->declare('v1.1.0');
 
 use Mail::Milter::Authentication;
 use Module::Load;
@@ -15,6 +15,8 @@ our @EXPORT_OK = qw{
 use JSON;
 
 our $PREFIX = '/etc';
+our $IDENT  = 'authentication_milter';
+my  $CONFIG;
 
 sub default_config {
     my $config = {
@@ -37,9 +39,7 @@ sub default_config {
         'command_timeout'                 => 30,
         'content_timeout'                 => 300,
         'dns_timeout'                     => 10,
-        'dns_cache_timeout'               => 240,
         'dns_retry'                       => 2,
-        'dns_cache_error_limit'           => 3,
         'tempfail_on_error'               => '1',
         'tempfail_on_error_authenticated' => '0',
         'tempfail_on_error_local'         => '0',
@@ -89,6 +89,8 @@ sub load_file {
 
 sub get_config {
 
+    return $CONFIG if $CONFIG;
+
     my $file = $PREFIX . '/authentication_milter.json';
 
     my $config = load_file( $file );
@@ -118,6 +120,8 @@ sub get_config {
     my $protocol = $config->{'protocol'} || 'milter';
     $config->{'protocol'} = $protocol;
 
+    $CONFIG = $config;
+
     return $config;
 
 }
