@@ -2,10 +2,11 @@ package Mail::Milter::Authentication::Handler::DMARC;
 use strict;
 use warnings;
 use base 'Mail::Milter::Authentication::Handler';
-use version; our $VERSION = version->declare('v1.0.1');
+use version; our $VERSION = version->declare('v1.1.0');
 
 use Data::Dumper;
 use English qw{ -no_match_vars };
+use Net::IP;
 use Sys::Syslog qw{:standard :macros};
 
 use Mail::DMARC::PurePerl;
@@ -15,12 +16,49 @@ my $PSL_CHECKED_TIME;
 sub default_config {
     return {
         'hide_none'      => 0,
+        'hard_reject'    => 0,
+        'no_list_reject' => 1,
+        'whitelisted'    => [],
         'detect_list_id' => 1,
         'report_skip_to' => [ 'my_report_from_address@example.com' ],
         'no_report'      => 0,
     };
 }
 
+sub is_whitelisted {
+    my ( $self ) = @_;
+    my $config = $self->handler_config();
+    return 0 if not exists( $config->{'whitelisted'} );
+    my $top_handler = $self->get_top_handler();
+    my $ip_obj = $top_handler->{'ip_object'};
+    my $whitelisted = 0;
+    foreach my $entry ( @{ $config->{'whitelisted'} } ) {
+        if ( $entry =~ /^dkim:/ ) {
+            my ( $dummy, $dkim_domain ) = split( /:/, $entry, 2 );
+            my $dkim_handler = $self->get_handler('DKIM');
+            if ( exists( $dkim_handler->{'valid_domains'}->{ lc $dkim_domain } ) ) {
+                $self->dbgout( 'DMARCReject', "Whitelist hit " . $entry, LOG_INFO );
+                $whitelisted = 1;
+            }
+        }
+        else {
+            my $whitelisted_obj = Net::IP->new($entry);
+            my $is_overlap = $ip_obj->overlaps($whitelisted_obj) || 0;
+            if (
+                   $is_overlap == $IP_A_IN_B_OVERLAP
+                || $is_overlap == $IP_B_IN_A_OVERLAP     # Should never happen
+                || $is_overlap == $IP_PARTIAL_OVERLAP    # Should never happen
+                || $is_overlap == $IP_IDENTICAL
+              )
+            {
+                $self->dbgout( 'DMARCReject', "Whitelist hit " . $entry, LOG_INFO );
+                $whitelisted = 1;
+            }
+        }
+    }
+    return $whitelisted;
+}
+
 sub pre_loop_setup {
     my ( $self ) = @_;
     $PSL_CHECKED_TIME = time;
@@ -83,7 +121,7 @@ sub get_dmarc_object {
         $self->add_auth_header('dmarc=temperror');
         $self->{'failmode'} = 1;
     }
-    
+
     return $dmarc;
 }
 
@@ -279,6 +317,7 @@ sub eom_callback {
         my $dmarc_result = $dmarc->validate();
         my $dmarc_code   = $dmarc_result->result;
         $self->dbgout( 'DMARCCode', $dmarc_code, LOG_INFO );
+
         if ( !( $config->{'hide_none'} && $dmarc_code eq 'none' ) ) {
             my $dmarc_policy;
             if ( $dmarc_code ne 'pass' ) {
@@ -287,6 +326,21 @@ sub eom_callback {
                     $self->log_error( 'DMARCPolicyError ' . $error );
                 }
                 $self->dbgout( 'DMARCPolicy', $dmarc_policy, LOG_INFO );
+                if ( $dmarc_code eq 'fail' && $dmarc_policy eq 'reject' ) {
+                    if ( $config->{'hard_reject'} ) {
+                        if ( $config->{'no_list_reject'} && $self->{'is_list'} ) {
+                            $self->dbgout( 'DMARCReject', "Policy reject overridden for list mail", LOG_INFO );
+                        }
+                        elsif ( $self->is_whitelisted() ) {
+                            $self->dbgout( 'DMARCReject', "Policy reject overridden by whitelist", LOG_INFO );
+                        }
+                        else {
+                            $self->reject_mail( '550 5.7.0 DMARC policy violation' );
+                            $self->dbgout( 'DMARCReject', "Policy reject", LOG_INFO );
+                        }
+                    }
+                }
+
             }
             my $dmarc_header = $self->format_header_entry( 'dmarc', $dmarc_code );
             my $is_list_entry = q{};
@@ -371,15 +425,22 @@ This handler requires the SPF and DKIM handlers to be installed and active.
 
         "DMARC" : {                                     | Config for the DMARC Module
                                                         | Requires DKIM and SPF
-            "hide_none"      : 0,                       | Hide auth line if the result is 'none'
-            "detect_list_id" : "1",                     | Detect a list ID and modify the DMARC authentication header
+            "hard_reject"         : 0,                  | Reject mail which fails with a reject policy
+            "no_list_reject"      : 0,                  | Do not reject mail detected as mailing list
+            "whitelisted"         : [                   | A list of ip addresses or CIDR ranges, or dkim domains
+                "10.20.30.40",                          | for which we do not want to hard reject mail on fail p=reject
+                "dkim:bad.forwarder.com",               | (valid) DKIM signing domains can also be whitelisted by
+                "20.30.40.0/24"                         | having an entry such as "dkim:domain.com"
+            ],
+            "hide_none"           : 0,                  | Hide auth line if the result is 'none'
+            "detect_list_id"      : "1",                | Detect a list ID and modify the DMARC authentication header
                                                         | to note this, useful when making rules for junking email
                                                         | as mailing lists frequently cause false DMARC failures.
-            "report_skip_to" : [                        | Do not send DMARC reports for emails to these addresses.
+            "report_skip_to"     : [                    | Do not send DMARC reports for emails to these addresses.
                 "dmarc@yourdomain.com",                 | This can be used to avoid report loops for email sent to
                 "dmarc@example.com"                     | your report from addresses.
             ],
-            "no_report" : "1"                           | If set then we will not attempt to store DMARC reports.
+            "no_report"          : "1"                  | If set then we will not attempt to store DMARC reports.
         },
 
 =head1 SYNOPSIS
