@@ -1,7 +1,7 @@
 package Mail::Milter::Authentication::Protocol::Milter;
 use strict;
 use warnings;
-use version; our $VERSION = version->declare('v1.0.1');
+use version; our $VERSION = version->declare('v1.1.0');
 
 use English qw{ -no_match_vars };
 use Net::IP;
@@ -17,7 +17,7 @@ sub protocol_process_request {
     COMMAND:
     while ( 1 ) {
 
-        # Get packet length 
+        # Get packet length
         my $length = unpack('N', $self->milter_read_block(4) ) || last;
         $self->fatal("bad packet length $length") if ($length <= 0 || $length > 131072);
 
@@ -34,7 +34,7 @@ sub protocol_process_request {
         last COMMAND if $command eq SMFIC_QUIT;
         $self->milter_process_command( $command, $data );
 
-    }    
+    }
 
     return;
 }
@@ -112,6 +112,12 @@ sub milter_process_command {
         $self->fatal("Unknown milter command $command");
     }
 
+    my $reject_reason;
+    if ( $reject_reason = $handler->get_reject_mail() ) {
+        $handler->clear_reject_mail();
+        $returncode = SMFIS_REJECT;
+    }
+
     if (defined $returncode) {
         if ( $returncode == SMFIS_CONTINUE ) {
             $returncode = SMFIR_CONTINUE;
@@ -138,10 +144,26 @@ sub milter_process_command {
         }
 
         if ( $command ne SMFIC_ABORT ) {
-            $self->write_packet($returncode);
+            if ( $reject_reason ) {
+                my ( $rcode, $xcode, $message ) = split( ' ', $reject_reason, 3 );
+                if ($rcode !~ /^[45]\d\d$/ || $xcode !~ /^[45]\.\d\.\d$/ || substr($rcode, 0, 1) ne substr($xcode, 0, 1)) {
+                    $self->loginfo ( "Invalid reject message $reject_reason - setting to TempFail" );
+                    $self->write_packet(SMFIR_TEMPFAIL );
+                }
+                else {
+                    $self->loginfo ( "SMTPReject: $reject_reason" );
+                    $self->write_packet( SMFIR_REPLYCODE,
+                        $reject_reason
+                        . "\0"
+                    );
+                }
+            }
+            else {
+                $self->write_packet($returncode);
+            }
         }
-    } 
- 
+    }
+
     return;
 }
 
@@ -188,7 +210,7 @@ sub milter_read_block {
     my ( $self, $len ) = @_;
     my $socket = $self->{'socket'};
     my $sofar = 0;
-    my $buffer = q{}; 
+    my $buffer = q{};
     while ($len > $sofar) {
         my $read = $socket->sysread($buffer, $len - $sofar, $sofar);
         last if (!defined($read) || $read <= 0); # EOF
