@@ -2,6 +2,7 @@ package AuthMilterTest;
 
 use strict;
 use warnings;
+use AuthMilterTestDNSCache;
 use Test::More;
 use Test::File::Contents;
 
@@ -9,7 +10,7 @@ use Cwd qw{ cwd };
 use IO::Socket::INET;
 use IO::Socket::UNIX;
 use Module::Load;
-        
+
 use Mail::Milter::Authentication;
 use Mail::Milter::Authentication::Client;
 use Mail::Milter::Authentication::Config;
@@ -20,10 +21,6 @@ my $base_dir = cwd();
 
 our $MASTER_PROCESS_PID = $$;
 
-sub set_lib {
-    return 'export PERL5LIB=' . $base_dir . '/lib';
-}
-
 sub tools_test {
 
     my $catargs = {
@@ -35,7 +32,7 @@ sub tools_test {
     unlink 'tmp/tools_test.sock';
     my $cat_pid = smtpcat( $catargs );
     sleep 2;
-    
+
     smtpput({
         'sock_type'    => 'unix',
         'sock_path'    => 'tmp/tools_test.sock',
@@ -66,7 +63,7 @@ sub tools_pipeline_test {
     unlink 'tmp/tools_test.sock';
     my $cat_pid = smtpcat( $catargs );
     sleep 2;
-    
+
     my $putargs = {
         'sock_type'    => 'unix',
         'sock_path'    => 'tmp/tools_test.sock',
@@ -122,15 +119,17 @@ sub tools_pipeline_test {
         if ( ! -e $prefix . '/authentication_milter.json' ) {
             die "Could not find config";
         }
-    
+
         system "cp $prefix/mail-dmarc.ini .";
-        
+
         $milter_pid = fork();
         die "unable to fork: $!" unless defined($milter_pid);
         if (!$milter_pid) {
             $Mail::Milter::Authentication::Config::PREFIX = $prefix;
+            $Mail::Milter::Authentication::Config::IDENT  = 'test_authentication_milter_test';
+            $Mail::Milter::Authentication::Handler::TestResolver = AuthMilterTestDNSCache->new(),
             Mail::Milter::Authentication::start({
-               'pid_file'   => 'tmp/authentication_milter.pid',
+                'pid_file'   => 'tmp/authentication_milter.pid',
                 'daemon'     => 0,
             });
             die;
@@ -178,10 +177,13 @@ sub smtp_process {
         'output'    => 'tmp/result/' . $args->{'dest'},
     };
     unlink 'tmp/authentication_milter_smtp_out.sock';
-    my $cat_pid = smtpcat( $catargs );
-    sleep 2;
+    my $cat_pid;
+    if ( ! $args->{'no_cat'} ) {
+        $cat_pid = smtpcat( $catargs );
+        sleep 2;
+    }
 
-    smtpput({
+    my $return = smtpput({
         'sock_type'    => 'unix',
         'sock_path'    => 'tmp/authentication_milter_test.sock',
         'mailer_name'  => 'test.module',
@@ -191,11 +193,16 @@ sub smtp_process {
         'mail_from'    => [ $args->{'from'} ],
         'rcpt_to'      => [ $args->{'to'} ],
         'mail_file'    => [ 'data/source/' . $args->{'source'} ],
+        'eom_expect'   => $args->{'eom_expect'},
     });
 
-    waitpid( $cat_pid,0 );
-
-    files_eq( 'data/example/' . $args->{'dest'}, 'tmp/result/' . $args->{'dest'}, 'smtp ' . $args->{'desc'} );
+    if ( ! $args->{'no_cat'} ) {
+        waitpid( $cat_pid,0 );
+        files_eq( 'data/example/' . $args->{'dest'}, 'tmp/result/' . $args->{'dest'}, 'smtp ' . $args->{'desc'} );
+    }
+    else {
+        is( $return, 1, 'SMTP Put Returned ok' );
+    }
 
     return;
 }
@@ -221,7 +228,7 @@ sub smtp_process_multi {
     unlink 'tmp/authentication_milter_smtp_out.sock';
     my $cat_pid = smtpcat( $catargs );
     sleep 2;
-    
+
     my $putargs = {
         'sock_type'    => 'unix',
         'sock_path'    => 'tmp/authentication_milter_test.sock',
@@ -348,7 +355,7 @@ sub run_milter_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     milter_process({
         'desc'   => 'SPF Fail',
         'prefix' => 'config/normal',
@@ -359,7 +366,7 @@ sub run_milter_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     milter_process({
         'desc'   => 'DKIM Fail Domain Space',
         'prefix' => 'config/normal',
@@ -381,7 +388,7 @@ sub run_milter_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     milter_process({
         'desc'   => 'DKIM/SPF Fail',
         'prefix' => 'config/normal',
@@ -392,7 +399,7 @@ sub run_milter_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     milter_process({
         'desc'   => 'No DKIM',
         'prefix' => 'config/normal',
@@ -403,7 +410,7 @@ sub run_milter_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     milter_process({
         'desc'   => 'No DKIM/SPF Fail',
         'prefix' => 'config/normal',
@@ -414,7 +421,7 @@ sub run_milter_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     milter_process({
         'desc'   => 'Sanitize Headers',
         'prefix' => 'config/normal',
@@ -425,7 +432,7 @@ sub run_milter_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     milter_process({
         'desc'   => 'Long Lines',
         'prefix' => 'config/normal',
@@ -436,11 +443,74 @@ sub run_milter_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
+    milter_process({
+        'desc'   => 'DMARC Reject',
+        'prefix' => 'config/normal',
+        'source' => 'dmarc_reject.eml',
+        'dest'   => 'dmarc_reject.eml',
+        'ip'     => '123.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'test@goestheweasel.com',
+        'to'     => 'marc@fastmail.com',
+    });
+
     stop_milter();
-    
-    start_milter( 'config/dryrun' ); 
-    
+
+    start_milter( 'config/dmarc_reject' );
+
+    milter_process({
+        'desc'   => 'DMARC Reject Hard',
+        'prefix' => 'config/dmarc_reject',
+        'source' => 'dmarc_reject.eml',
+        'dest'   => 'dmarc_reject_hard.eml',
+        'ip'     => '123.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'test@goestheweasel.com',
+        'to'     => 'marc@fastmail.com',
+    });
+
+    milter_process({
+        'desc'   => 'DKIM/SPF Fail Hard',
+        'prefix' => 'config/dmarc_reject',
+        'source' => 'google_apps_bad.eml',
+        'dest'   => 'google_apps_bad_spf_fail_hard.eml',
+        'ip'     => '123.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'marc@marcbradshaw.net',
+        'to'     => 'marc@fastmail.com',
+    });
+
+    milter_process({
+        'desc'   => 'DMARC Reject Hard Whitelisted',
+        'prefix' => 'config/dmarc_reject',
+        'source' => 'dmarc_reject.eml',
+        'dest'   => 'dmarc_reject_hard_whitelisted.eml',
+        'ip'     => '99.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'test@goestheweasel.com',
+        'to'     => 'marc@fastmail.com',
+    });
+
+    stop_milter();
+
+    start_milter( 'config/dmarc_reject_wl' );
+
+    milter_process({
+        'desc'   => 'DMARC Reject DKIM Whitelisted',
+        'prefix' => 'config/dmarc_reject_wl',
+        'source' => 'dmarc_reject_wl.eml',
+        'dest'   => 'dmarc_reject_wl.eml',
+        'ip'     => '123.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'marc@marcbradshaw.net',
+        'to'     => 'marc@fastmail.com',
+    });
+
+    stop_milter();
+
+    start_milter( 'config/dryrun' );
+
     milter_process({
         'desc'   => 'Dry Run Mode',
         'prefix' => 'config/normal',
@@ -451,7 +521,7 @@ sub run_milter_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     stop_milter();
 
     return;
@@ -490,7 +560,7 @@ sub run_smtp_processing {
         'name'   => [ 'test.example.com', 'localhost', 'bad.name.google.com', ],
         'from'   => [ 'test@example.com', 'marc@marcbradshaw.net', 'marc@marcbradshaw.net', ],
         'to'     => [ 'test@example.com', 'marc@fastmail.com', 'marc@fastmail.com', ],
-        'filter' => [10,11,51,52,121,122],
+        'filter' => [10,11,52,53,122,123],
     });
 
     smtp_process_multi({
@@ -502,7 +572,7 @@ sub run_smtp_processing {
         'name'   => [ 'test.example.com', 'localhost', 'bad.name.google.com', 'test.example.com', 'localhost', 'bad.name.google.com', ],
         'from'   => [ 'test@example.com', 'marc@marcbradshaw.net', 'marc@marcbradshaw.net', 'test@example.com', 'marc@marcbradshaw.net', 'marc@marcbradshaw.net', ],
         'to'     => [ 'test@example.com', 'marc@fastmail.com', 'marc@fastmail.com', 'test@example.com', 'marc@fastmail.com', 'marc@fastmail.com', ],
-        'filter' => [10,11,51,52,121,122,202,203],
+        'filter' => [10,11,52,53,122,123,203,204],
     });
 
     smtp_process({
@@ -515,7 +585,7 @@ sub run_smtp_processing {
         'from'   => 'test@example.com',
         'to'     => 'test@example.com',
     });
-    
+
     smtp_process({
         'desc'   => '8BITMIME message',
         'prefix' => 'config/normal.smtp',
@@ -581,7 +651,7 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     smtp_process({
         'desc'   => 'Good message no from',
         'prefix' => 'config/normal.smtp',
@@ -603,7 +673,7 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     smtp_process({
         'desc'   => 'SPF Fail',
         'prefix' => 'config/normal.smtp',
@@ -614,7 +684,7 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     smtp_process({
         'desc'   => 'DKIM Fail Domain Space',
         'prefix' => 'config/normal.smtp',
@@ -635,7 +705,7 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     smtp_process({
         'desc'   => 'DKIM/SPF Fail',
         'prefix' => 'config/normal.smtp',
@@ -646,7 +716,7 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     smtp_process({
         'desc'   => 'No DKIM',
         'prefix' => 'config/normal.smtp',
@@ -657,7 +727,7 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     smtp_process({
         'desc'   => 'No DKIM/SPF Fail',
         'prefix' => 'config/normal.smtp',
@@ -668,7 +738,7 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     smtp_process({
         'desc'   => 'Sanitize Headers',
         'prefix' => 'config/normal.smtp',
@@ -679,7 +749,7 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     smtp_process({
         'desc'   => 'Long Lines',
         'prefix' => 'config/normal.smtp',
@@ -690,11 +760,76 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
+    smtp_process({
+        'desc'   => 'DMARC Reject',
+        'prefix' => 'config/normal.smtp',
+        'source' => 'dmarc_reject.eml',
+        'dest'   => 'dmarc_reject.smtp.eml',
+        'ip'     => '123.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'test@goestheweasel.com',
+        'to'     => 'marc@fastmail.com',
+    });
+
     stop_milter();
-    
+
+    start_milter( 'config/dmarc_reject.smtp' );
+
+    smtp_process({
+        'desc'   => 'DMARC Reject Hard',
+        'prefix' => 'config/dmarc_reject.smtp',
+        'source' => 'dmarc_reject.eml',
+        'dest'   => 'dmarc_reject_hard.smtp.eml',
+        'ip'     => '123.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'test@goestheweasel.com',
+        'to'     => 'marc@fastmail.com',
+        'eom_expect' => '550',
+        'no_cat' => 1,
+    });
+
+    smtp_process({
+        'desc'   => 'DKIM/SPF Fail Hard',
+        'prefix' => 'config/dmarc_reject.smtp',
+        'source' => 'google_apps_bad.eml',
+        'dest'   => 'google_apps_bad_spf_fail_hard.smtp.eml',
+        'ip'     => '123.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'marc@marcbradshaw.net',
+        'to'     => 'marc@fastmail.com',
+    });
+
+    smtp_process({
+        'desc'   => 'DMARC Reject Hard Whitelisted',
+        'prefix' => 'config/dmarc_reject.smtp',
+        'source' => 'dmarc_reject.eml',
+        'dest'   => 'dmarc_reject_hard_whitelisted.smtp.eml',
+        'ip'     => '99.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'test@goestheweasel.com',
+        'to'     => 'marc@fastmail.com',
+    });
+
+    stop_milter();
+
+    start_milter( 'config/dmarc_reject_wl.smtp' );
+
+    smtp_process({
+        'desc'   => 'DMARC Reject DKIM Whitelisted',
+        'prefix' => 'config/dmarc_reject_wl.smtp',
+        'source' => 'dmarc_reject_wl.eml',
+        'dest'   => 'dmarc_reject_wl.smtp.eml',
+        'ip'     => '123.123.123.123',
+        'name'   => 'bad.name.google.com',
+        'from'   => 'marc@marcbradshaw.net',
+        'to'     => 'marc@fastmail.com',
+    });
+
+    stop_milter();
+
     start_milter( 'config/dryrun.smtp' );
-    
+
     smtp_process({
         'desc'   => 'Dry Run Mode',
         'prefix' => 'config/normal',
@@ -705,7 +840,7 @@ sub run_smtp_processing {
         'from'   => 'marc@marcbradshaw.net',
         'to'     => 'marc@fastmail.com',
     });
-    
+
     stop_milter();
 
     return;
@@ -735,19 +870,21 @@ sub smtpput {
     my ( $args ) = @_;
 
     my $mailer_name  = $args->{'mailer_name'};
-    
+
     my $mail_file_a  = $args->{'mail_file'};
     my $mail_from_a  = $args->{'mail_from'};
     my $rcpt_to_a    = $args->{'rcpt_to'};
     my $x_name_a     = $args->{'connect_name'};
     my $x_addr_a     = $args->{'connect_ip'};
     my $x_helo_a     = $args->{'helo_host'};
-    
+
     my $sock_type    = $args->{'sock_type'};
     my $sock_path    = $args->{'sock_path'};
     my $sock_host    = $args->{'sock_host'};
     my $sock_port    = $args->{'sock_port'};
-    
+
+    my $eom_expect   = $args->{'eom_expect'} || '250';
+
     my $sock;
     if ( $sock_type eq 'inet' ) {
        $sock = IO::Socket::INET->new(
@@ -761,20 +898,20 @@ sub smtpput {
             'Peer' => $sock_path,
         ) || die "could not open outbound SMTP socket: $!";
     }
-    
+
     my $line = <$sock>;
-    
+
     if ( ! $line =~ /250/ ) {
         die "Unexpected SMTP response $line";
         return 0;
     }
-    
+
     send_smtp_packet( $sock, 'EHLO ' . $mailer_name,       '250' ) || die;
-    
+
     my $first_time = 1;
-   
+
     while ( @$mail_from_a ) {
-    
+
         if ( ! $first_time ) {
             if ( ! send_smtp_packet( $sock, 'RSET', '250' ) ) {
                 $sock->close();
@@ -782,16 +919,16 @@ sub smtpput {
             };
         }
         $first_time = 0;
-    
+
         my $mail_file = shift @$mail_file_a;
-        my $mail_from = shift @$mail_from_a; 
+        my $mail_from = shift @$mail_from_a;
         my $rcpt_to   = shift @$rcpt_to_a;
         my $x_name    = shift @$x_name_a;
         my $x_addr    = shift @$x_addr_a;
         my $x_helo    = shift @$x_helo_a;
-        
+
         my $mail_data = q{};
-        
+
         if ( $mail_file eq '-' ) {
             while ( my $l = <> ) {
                 $mail_data .= $l;
@@ -806,30 +943,30 @@ sub smtpput {
             $mail_data = join( q{}, @all );
             close $inf;
         }
-        
+
         $mail_data =~ s/\015?\012/\015\012/g;
         # Handle transparency
         $mail_data =~ s/\015\012\./\015\012\.\./g;
-        
+
         send_smtp_packet( $sock, 'XFORWARD NAME=' . $x_name,   '250' ) || die;
         send_smtp_packet( $sock, 'XFORWARD ADDR=' . $x_addr,   '250' ) || die;
         send_smtp_packet( $sock, 'XFORWARD HELO=' . $x_helo,   '250' ) || die;
-        
+
         send_smtp_packet( $sock, 'MAIL FROM:' . $mail_from, '250' ) || die;
         send_smtp_packet( $sock, 'RCPT TO:' .   $rcpt_to,   '250' ) || die;
         send_smtp_packet( $sock, 'DATA',                    '354' ) || die;
-        
+
         print $sock $mail_data;
         print $sock "\r\n";
-        
-        send_smtp_packet( $sock, '.',    '250' ) || return;
-    
+
+        send_smtp_packet( $sock, '.',    $eom_expect ) || return 0;
+
     }
-        
-    send_smtp_packet( $sock, 'QUIT', '221' ) || return;
+
+    send_smtp_packet( $sock, 'QUIT', '221' ) || return 0;
     $sock->close();
 
-    return;
+    return 1;
 }
 
 sub send_smtp_packet {
@@ -850,19 +987,19 @@ sub send_smtp_packet {
 
 sub smtpcat {
     my ( $args ) = @_;
-    
+
     my $cat_pid = fork();
     die "unable to fork: $!" unless defined($cat_pid);
-    return $cat_pid if $cat_pid; 
-    
+    return $cat_pid if $cat_pid;
+
     my $sock_type = $args->{'sock_type'};
     my $sock_path = $args->{'sock_path'};
     my $sock_host = $args->{'sock_host'};
     my $sock_port = $args->{'sock_port'};
-    
+
     my $remove = $args->{'remove'};
     my $output = $args->{'output'};
-   
+
     my @out_lines;
 
     my $sock;
@@ -880,19 +1017,19 @@ sub smtpcat {
             'Local' => $sock_path,
         ) || die "could not open socket: $!";
     }
-    
+
     my $accept = $sock->accept();
-    
+
     print $accept "220 smtp.cat ESMTP Test\r\n";
-    
+
     local $SIG{'ALRM'} = sub{ die "Timeout\n" };
     alarm( 60 );
-    
+
     my $quit = 0;
     while ( ! $quit ) {
         my $command = <$accept> || { $quit = 1 };
         alarm( 60 );
-    
+
         if ( $command =~ /^HELO/ ) {
             push @out_lines, $command;
             print $accept "250 HELO Ok\r\n";
@@ -969,6 +1106,7 @@ sub client {
 
         $Mail::Milter::Authentication::Config::PREFIX = $args->{'prefix'};
         delete $args->{'prefix'};
+        $args->{'testing'} = 1;
 
         my $client = Mail::Milter::Authentication::Client->new( $args );
 
